package com.sun.jmx.mbeanserver;

import java.lang.reflect.Proxy;
import javax.management.openmbean.CompositeDataInvocationHandler;
import java.lang.reflect.InvocationTargetException;
import java.lang.annotation.Annotation;
import java.util.BitSet;
import java.lang.reflect.Constructor;
import java.lang.reflect.Modifier;
import sun.reflect.misc.MethodUtil;
import javax.management.openmbean.CompositeDataView;
import javax.management.openmbean.CompositeData;
import javax.management.openmbean.TabularData;
import javax.management.openmbean.CompositeDataSupport;
import javax.management.openmbean.TabularDataSupport;
import java.util.Comparator;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.TreeSet;
import java.util.Collection;
import java.util.WeakHashMap;
import java.lang.reflect.Array;
import javax.management.ObjectName;
import javax.management.openmbean.SimpleType;
import com.sun.jmx.remote.util.EnvHelp;
import java.io.InvalidObjectException;
import java.util.Iterator;
import java.lang.reflect.Method;
import java.util.Arrays;
import java.util.SortedMap;
import java.util.SortedSet;
import java.util.Set;
import javax.management.openmbean.TabularType;
import javax.management.openmbean.CompositeType;
import javax.management.openmbean.OpenType;
import javax.management.openmbean.ArrayType;
import sun.reflect.misc.ReflectUtil;
import java.lang.annotation.ElementType;
import javax.management.JMX;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.GenericArrayType;
import javax.management.openmbean.OpenDataException;
import java.lang.ref.WeakReference;
import java.lang.reflect.Type;
import java.util.Map;
import java.util.List;

public class DefaultMXBeanMappingFactory extends MXBeanMappingFactory
{
    private static final DefaultMXBeanMappingFactory.Mappings mappings;
    private static final List<MXBeanMapping> permanentMappings;
    private static final String[] keyArray;
    private static final String[] keyValueArray;
    private static final Map<Type, Type> inProgress;
    
    public DefaultMXBeanMappingFactory() {
    }
    
    static boolean isIdentity(MXBeanMapping mxBeanMapping) {
        return mxBeanMapping instanceof DefaultMXBeanMappingFactory.NonNullMXBeanMapping && ((DefaultMXBeanMappingFactory.NonNullMXBeanMapping)mxBeanMapping).isIdentity();
    }
    
    private static synchronized MXBeanMapping getMapping(Type key) {
        WeakReference<Object> weakReference = (WeakReference)DefaultMXBeanMappingFactory.mappings.get(key);
        return (weakReference == null) ? null : ((MXBeanMapping)weakReference.get());
    }
    
    private static synchronized void putMapping(Type key, MXBeanMapping referent) {
        DefaultMXBeanMappingFactory.mappings.put(key, new WeakReference(referent));
    }
    
    private static synchronized void putPermanentMapping(Type type, MXBeanMapping mxBeanMapping) {
        putMapping(type, mxBeanMapping);
        DefaultMXBeanMappingFactory.permanentMappings.add(mxBeanMapping);
    }
    
    @Override
    public synchronized MXBeanMapping mappingForType(Type type, MXBeanMappingFactory mxBeanMappingFactory) throws OpenDataException {
        if (DefaultMXBeanMappingFactory.inProgress.containsKey(type)) {
            throw new OpenDataException("Recursive data structure, including " + MXBeanIntrospector.typeName(type));
        }
        MXBeanMapping mxBeanMapping = getMapping(type);
        if (mxBeanMapping != null) {
            return mxBeanMapping;
        }
        DefaultMXBeanMappingFactory.inProgress.put(type, type);
        try {
            mxBeanMapping = this.makeMapping(type, mxBeanMappingFactory);
        }
        catch (OpenDataException ex) {
            throw openDataException("Cannot convert type: " + MXBeanIntrospector.typeName(type), ex);
        }
        finally {
            DefaultMXBeanMappingFactory.inProgress.remove(type);
        }
        putMapping(type, mxBeanMapping);
        return mxBeanMapping;
    }
    
    private MXBeanMapping makeMapping(Type obj, MXBeanMappingFactory mxBeanMappingFactory) throws OpenDataException {
        if ((obj instanceof GenericArrayType)) {
            return this.makeArrayOrCollectionMapping(obj, ((GenericArrayType)obj).getGenericComponentType(), mxBeanMappingFactory);
        }
        if ((obj instanceof Class)) {
            Class<Object> interfaceClass = (Class)obj;
            if (interfaceClass.isEnum()) {
                return makeEnumMapping(interfaceClass, ElementType.class);
            }
            if (interfaceClass.isArray()) {
                return this.makeArrayOrCollectionMapping(interfaceClass, interfaceClass.getComponentType(), mxBeanMappingFactory);
            }
            if (JMX.isMXBeanInterface(interfaceClass)) {
                return makeMXBeanRefMapping(interfaceClass);
            }
            return this.makeCompositeMapping(interfaceClass, mxBeanMappingFactory);
        }
        else {
            if ((obj instanceof ParameterizedType)) {
                return this.makeParameterizedTypeMapping((ParameterizedType)obj, mxBeanMappingFactory);
            }
            throw new OpenDataException("Cannot map type: " + obj);
        }
    }
    
    private static <T extends Enum<T>> MXBeanMapping makeEnumMapping(Class<?> clazz, Class<T> clazz2) {
        ReflectUtil.checkPackageAccess(clazz);
        return new DefaultMXBeanMappingFactory.EnumMapping((Class)Util.cast(clazz));
    }
    
    private MXBeanMapping makeArrayOrCollectionMapping(Type type, Type type2, MXBeanMappingFactory mxBeanMappingFactory) throws OpenDataException {
        MXBeanMapping mappingForType = mxBeanMappingFactory.mappingForType(type2, mxBeanMappingFactory);
        ArrayType<Object> arrayType = ArrayType.getArrayType(mappingForType.getOpenType());
        Class<Object> openClass = mappingForType.getOpenClass();
        String className;
        if (openClass.isArray()) {
            className = "[" + openClass.getName();
        }
        else {
            className = "[L" + openClass.getName() + ";";
        }
        Class<Object> forName;
        try {
            forName = Class.forName(className);
        }
        catch (ClassNotFoundException ex) {
            throw openDataException("Cannot obtain array class", ex);
        }
        if ((type instanceof ParameterizedType)) {
            return new DefaultMXBeanMappingFactory.CollectionMapping(type, arrayType, forName, mappingForType);
        }
        if (isIdentity(mappingForType)) {
            return new DefaultMXBeanMappingFactory.IdentityMapping(type, arrayType);
        }
        return new DefaultMXBeanMappingFactory.ArrayMapping(type, arrayType, forName, mappingForType);
    }
    
    private MXBeanMapping makeTabularMapping(Type type, boolean b, Type type2, Type type3, MXBeanMappingFactory mxBeanMappingFactory) throws OpenDataException {
        String typeName = MXBeanIntrospector.typeName(type);
        MXBeanMapping mappingForType = mxBeanMappingFactory.mappingForType(type2, mxBeanMappingFactory);
        MXBeanMapping mappingForType2 = mxBeanMappingFactory.mappingForType(type3, mxBeanMappingFactory);
        return new DefaultMXBeanMappingFactory.TabularMapping(type, b, new TabularType(typeName, typeName, new CompositeType(typeName, typeName, DefaultMXBeanMappingFactory.keyValueArray, DefaultMXBeanMappingFactory.keyValueArray, new OpenType[] { mappingForType.getOpenType(), mappingForType2.getOpenType() }), DefaultMXBeanMappingFactory.keyArray), mappingForType, mappingForType2);
    }
    
    private MXBeanMapping makeParameterizedTypeMapping(ParameterizedType obj, MXBeanMappingFactory mxBeanMappingFactory) throws OpenDataException {
        Type rawType = obj.getRawType();
        if ((rawType instanceof Class)) {
            Class<Object> clazz = (Class)rawType;
            if (clazz == List.class || clazz == Set.class || clazz == SortedSet.class) {
                Type[] actualTypeArguments2 = obj.getActualTypeArguments();
                assert actualTypeArguments2.length == 1;
                if (clazz == SortedSet.class) {
                    mustBeComparable(clazz, actualTypeArguments2[0]);
                }
                return this.makeArrayOrCollectionMapping(obj, actualTypeArguments2[0], mxBeanMappingFactory);
            }
            else {
                int n = clazz == SortedMap.class;
                if (clazz == Map.class || n) {
                    Type[] actualTypeArguments = obj.getActualTypeArguments();
                    assert actualTypeArguments.length == 2;
                    if (!(n == 0)) {
                        mustBeComparable(clazz, actualTypeArguments[0]);
                    }
                    return this.makeTabularMapping(obj, (boolean)n, actualTypeArguments[0], actualTypeArguments[1], mxBeanMappingFactory);
                }
            }
        }
        throw new OpenDataException("Cannot convert type: " + obj);
    }
    
    private static MXBeanMapping makeMXBeanRefMapping(Type type) throws OpenDataException {
        return new MXBeanRefMapping(type);
    }
    
    private MXBeanMapping makeCompositeMapping(Class<?> clazz, MXBeanMappingFactory mxBeanMappingFactory) throws OpenDataException {
        int n = clazz.getName().equals("com.sun.management.GcInfo") && clazz.getClassLoader() == null;
        ReflectUtil.checkPackageAccess(clazz);
        List<Method> eliminateCovariantMethods = MBeanAnalyzer.eliminateCovariantMethods(Arrays.asList(clazz.getMethods()));
        SortedMap<String, Method> sortedMap = Util.newSortedMap();
        for (Method method : eliminateCovariantMethods) {
            String propertyName = propertyName(method);
            if (propertyName == null) {
                continue;
            }
            if (n && propertyName.equals("CompositeType")) {
                continue;
            }
            Method method2 = (Method)sortedMap.put(decapitalize(propertyName), method);
            if (method2 != null) {
                throw new OpenDataException("Class " + clazz.getName() + " has method name clash: " + method2.getName() + ", " + method.getName());
            }
            continue;
        }
        int size = sortedMap.size();
        if (size == 0) {
            throw new OpenDataException("Can't map " + clazz.getName() + " to an open data type");
        }
        Method[] array = new Method[size];
        String[] array2 = new String[size];
        OpenType[] itemTypes = new OpenType[size];
        int n2 = 0;
        for (Map.Entry entry : sortedMap.entrySet()) {
            array2[n2] = (String)entry.getKey();
            Method method3 = (Method)entry.getValue();
            array[n2] = method3;
            itemTypes[n2] = mxBeanMappingFactory.mappingForType(method3.getGenericReturnType(), mxBeanMappingFactory).getOpenType();
            ++n2;
        }
        return new DefaultMXBeanMappingFactory.CompositeMapping(clazz, new CompositeType(clazz.getName(), clazz.getName(), array2, array2, itemTypes), array2, array, mxBeanMappingFactory);
    }
    
    static InvalidObjectException invalidObjectException(String reason, Throwable cause) {
        return (InvalidObjectException)EnvHelp.initCause(new InvalidObjectException(reason), cause);
    }
    
    static InvalidObjectException invalidObjectException(Throwable t) {
        return invalidObjectException(t.getMessage(), t);
    }
    
    static OpenDataException openDataException(String msg, Throwable cause) {
        return (OpenDataException)EnvHelp.initCause(new OpenDataException(msg), cause);
    }
    
    static OpenDataException openDataException(Throwable t) {
        return openDataException(t.getMessage(), t);
    }
    
    static void mustBeComparable(Class<?> clazz, Type obj) throws OpenDataException {
        if (!(obj instanceof Class) || !Comparable.class.isAssignableFrom((Class)obj)) {
            throw new OpenDataException("Parameter class " + obj + " of " + clazz.getName() + " does not implement " + Comparable.class.getName());
        }
        return;
    }
    
    public static String decapitalize(String seq) {
        if (seq == null || seq.length() == 0) {
            return seq;
        }
        int offsetByCodePoints = Character.offsetByCodePoints(seq, 0, 1);
        if (offsetByCodePoints < seq.length() && Character.isUpperCase(seq.codePointAt(offsetByCodePoints))) {
            return seq;
        }
        return seq.substring(0, offsetByCodePoints).toLowerCase() + seq.substring(offsetByCodePoints);
    }
    
    static String capitalize(String s) {
        if (s == null || s.length() == 0) {
            return s;
        }
        int offsetByCodePoints = s.offsetByCodePoints(0, 1);
        return s.substring(0, offsetByCodePoints).toUpperCase() + s.substring(offsetByCodePoints);
    }
    
    public static String propertyName(Method method) {
        String s = null;
        String name = method.getName();
        if (name.startsWith("get")) {
            s = name.substring(3);
        }
        else if (name.startsWith("is") && method.getReturnType() == Boolean.TYPE) {
            s = name.substring(2);
        }
        if (s == null || s.length() == 0 || method.getParameterTypes().length > 0 || method.getReturnType() == Void.TYPE || name.equals("getClass")) {
            return null;
        }
        return s;
    }
    
    static {
        mappings = new DefaultMXBeanMappingFactory.Mappings();
        permanentMappings = Util.newList();
        OpenType[] array = { SimpleType.BIGDECIMAL, SimpleType.BIGINTEGER, SimpleType.BOOLEAN, SimpleType.BYTE, SimpleType.CHARACTER, SimpleType.DATE, SimpleType.DOUBLE, SimpleType.FLOAT, SimpleType.INTEGER, SimpleType.LONG, SimpleType.OBJECTNAME, SimpleType.SHORT, SimpleType.STRING, SimpleType.VOID };
        for (int i = 0; i < array.length; ++i) {
            OpenType<Object> openType = array[i];
            Class<Object> forName;
            try {
                forName = Class.forName(openType.getClassName(), false, ObjectName.class.getClassLoader());
            }
            catch (ClassNotFoundException cause) {
                throw new Error(cause);
            }
            putPermanentMapping(forName, new DefaultMXBeanMappingFactory.IdentityMapping(forName, openType));
            if (forName.getName().startsWith("java.lang.")) {
                try {
                    Class<Object> componentType = (Class)forName.getField("TYPE").get(null);
                    putPermanentMapping(componentType, new DefaultMXBeanMappingFactory.IdentityMapping(componentType, openType));
                    if (componentType != Void.TYPE) {
                        Class<Object> class1 = Array.newInstance(componentType, 0).getClass();
                        putPermanentMapping(class1, new DefaultMXBeanMappingFactory.IdentityMapping(class1, ArrayType.getPrimitiveArrayType(class1)));
                    }
                }
                catch (NoSuchFieldException ex) {}
                catch (IllegalAccessException ex2) {
                    assert false;
                }
            }
        }
        keyArray = new String[] { "key" };
        keyValueArray = new String[] { "key", "value" };
        inProgress = Util.newIdentityHashMap();
    }
    
    abstract static class NonNullMXBeanMapping extends MXBeanMapping
    {
        NonNullMXBeanMapping(Type type, OpenType<?> openType) {
            super(type, openType);
        }
        
        @Override
        public final Object fromOpenValue(Object o) throws InvalidObjectException {
            if (o == null) {
                return null;
            }
            return this.fromNonNullOpenValue(o);
        }
        
        @Override
        public final Object toOpenValue(Object o) throws OpenDataException {
            if (o == null) {
                return null;
            }
            return this.toNonNullOpenValue(o);
        }
        
        abstract Object fromNonNullOpenValue(Object p0) throws InvalidObjectException;
        
        abstract Object toNonNullOpenValue(Object p0) throws OpenDataException;
        
        boolean isIdentity() {
            return false;
        }
    }
    
    private static final class Mappings extends WeakHashMap<Type, WeakReference<MXBeanMapping>>
    {
    }
    
    private static final class IdentityMapping extends DefaultMXBeanMappingFactory.NonNullMXBeanMapping
    {
        IdentityMapping(Type type, OpenType<?> openType) {
            super(type, openType);
        }
        
        @Override
        boolean isIdentity() {
            return true;
        }
        
        @Override
        Object fromNonNullOpenValue(Object o) throws InvalidObjectException {
            return o;
        }
        
        @Override
        Object toNonNullOpenValue(Object o) throws OpenDataException {
            return o;
        }
    }
    
    private static final class EnumMapping<T extends Enum<T>> extends DefaultMXBeanMappingFactory.NonNullMXBeanMapping
    {
        private final Class<T> enumClass;
        
        EnumMapping(Class<T> enumClass) {
            super(enumClass, SimpleType.STRING);
            this.enumClass = enumClass;
        }
        
        @Override
        final Object toNonNullOpenValue(Object o) {
            return ((Enum)o).name();
        }
        
        @Override
        final T fromNonNullOpenValue(Object obj) throws InvalidObjectException {
            try {
                return Enum.valueOf(this.enumClass, (String)obj);
            }
            catch (Exception ex) {
                throw DefaultMXBeanMappingFactory.invalidObjectException("Cannot convert to enum: " + obj, ex);
            }
        }
    }
    
    private static final class ArrayMapping extends DefaultMXBeanMappingFactory.NonNullMXBeanMapping
    {
        private final MXBeanMapping elementMapping;
        
        ArrayMapping(Type type, ArrayType<?> arrayType, Class<?> clazz, MXBeanMapping elementMapping) {
            super(type, arrayType);
            this.elementMapping = elementMapping;
        }
        
        @Override
        final Object toNonNullOpenValue(Object o) throws OpenDataException {
            Object[] array = (Object[])(Object[])o;
            int length = array.length;
            Object[] array2 = (Object[])(Object[])Array.newInstance(this.getOpenClass().getComponentType(), length);
            for (int i = 0; i < length; ++i) {
                array2[i] = this.elementMapping.toOpenValue(array[i]);
            }
            return array2;
        }
        
        @Override
        final Object fromNonNullOpenValue(Object o) throws InvalidObjectException {
            Object[] array = (Object[])(Object[])o;
            Type javaType = this.getJavaType();
            Type type;
            if ((javaType instanceof GenericArrayType)) {
                type = ((GenericArrayType)javaType).getGenericComponentType();
            }
            else if (javaType instanceof Class && ((Class)javaType).isArray()) {
                type = ((Class)javaType).getComponentType();
            }
            else {
                throw new IllegalArgumentException("Not an array: " + javaType);
            }
            Object[] array2 = (Object[])(Object[])Array.newInstance((Class)type, array.length);
            for (int i = 0; i < array.length; ++i) {
                array2[i] = this.elementMapping.fromOpenValue(array[i]);
            }
            return array2;
        }
        
        @Override
        public void checkReconstructible() throws InvalidObjectException {
            this.elementMapping.checkReconstructible();
        }
    }
    
    private static final class CollectionMapping extends DefaultMXBeanMappingFactory.NonNullMXBeanMapping
    {
        private final Class<? extends Collection<?>> collectionClass;
        private final MXBeanMapping elementMapping;
        
        CollectionMapping(Type type, ArrayType<?> arrayType, Class<?> clazz, MXBeanMapping elementMapping) {
            super(type, arrayType);
            this.elementMapping = elementMapping;
            Class<Object> clazz2 = (Class)((ParameterizedType)type).getRawType();
            Class<Object> clazz3;
            if (clazz2 == List.class) {
                clazz3 = ArrayList.class;
            }
            else if (clazz2 == Set.class) {
                clazz3 = HashSet.class;
            }
            else if (clazz2 == SortedSet.class) {
                clazz3 = TreeSet.class;
            }
            else {
                assert false;
                clazz3 = null;
            }
            this.collectionClass = (Class)Util.cast(clazz3);
        }
        
        @Override
        final Object toNonNullOpenValue(Object o) throws OpenDataException {
            Collection<Object> collection = (Collection)o;
            if ((collection instanceof SortedSet)) {
                Comparator<Object> comparator = ((SortedSet)collection).comparator();
                if (comparator != null) {
                    String string = "Cannot convert SortedSet with non-null comparator: " + comparator;
                    throw DefaultMXBeanMappingFactory.openDataException(string, new IllegalArgumentException(string));
                }
            }
            Object[] array = (Object[])(Object[])Array.newInstance(this.getOpenClass().getComponentType(), collection.size());
            int n = 0;
            Iterator<Object> iterator = collection.iterator();
            while (iterator.hasNext()) {
                array[n++] = this.elementMapping.toOpenValue(iterator.next());
            }
            return array;
        }
        
        @Override
        final Object fromNonNullOpenValue(Object o) throws InvalidObjectException {
            Object[] array = (Object[])(Object[])o;
            Collection<Object> collection;
            try {
                collection = (Collection)Util.cast(this.collectionClass.newInstance());
            }
            catch (Exception ex) {
                throw DefaultMXBeanMappingFactory.invalidObjectException("Cannot create collection", ex);
            }
            int length = array.length;
            int i = 0;
            while (i < length) {
                Object obj = array[i];
                if (!collection.add(this.elementMapping.fromOpenValue(obj))) {
                    throw new InvalidObjectException("Could not add " + obj + " to " + this.collectionClass.getName() + " (duplicate set element?)");
                }
                ++i;
            }
            return collection;
        }
        
        @Override
        public void checkReconstructible() throws InvalidObjectException {
            this.elementMapping.checkReconstructible();
        }
    }
    
    private static final class MXBeanRefMapping extends DefaultMXBeanMappingFactory.NonNullMXBeanMapping
    {
        MXBeanRefMapping(Type type) {
            super(type, SimpleType.OBJECTNAME);
        }
        
        @Override
        final Object toNonNullOpenValue(Object obj) throws OpenDataException {
            ObjectName mxbeanToObjectName = this.lookupNotNull(OpenDataException.class).mxbeanToObjectName(obj);
            if (mxbeanToObjectName == null) {
                throw new OpenDataException("No name for object: " + obj);
            }
            return mxbeanToObjectName;
        }
        
        @Override
        final Object fromNonNullOpenValue(Object o) throws InvalidObjectException {
            MXBeanLookup lookupNotNull = this.lookupNotNull(InvalidObjectException.class);
            ObjectName obj = (ObjectName)o;
            Object objectNameToMXBean = lookupNotNull.objectNameToMXBean(obj, (Class)this.getJavaType());
            if (objectNameToMXBean == null) {
                throw new InvalidObjectException("No MXBean for name: " + obj);
            }
            return objectNameToMXBean;
        }
        
        private <T extends Exception> MXBeanLookup lookupNotNull(Class<T> clazz) throws T, Exception {
            MXBeanLookup lookup = MXBeanLookup.getLookup();
            if (lookup == null) {
                Exception ex;
                try {
                    ex = (T)clazz.getConstructor(new Class[] { String.class }).newInstance(new Object[] { "Cannot convert MXBean interface in this context" });
                }
                catch (Exception cause) {
                    throw new RuntimeException(cause);
                }
                throw ex;
            }
            return lookup;
        }
    }
    
    private static final class TabularMapping extends DefaultMXBeanMappingFactory.NonNullMXBeanMapping
    {
        private final boolean sortedMap;
        private final MXBeanMapping keyMapping;
        private final MXBeanMapping valueMapping;
        
        TabularMapping(Type type, boolean sortedMap, TabularType tabularType, MXBeanMapping keyMapping, MXBeanMapping valueMapping) {
            super(type, tabularType);
            this.sortedMap = sortedMap;
            this.keyMapping = keyMapping;
            this.valueMapping = valueMapping;
        }
        
        @Override
        final Object toNonNullOpenValue(Object o) throws OpenDataException {
            Map<Object, Object> map = (Map)Util.cast(o);
            if ((map instanceof SortedMap)) {
                Comparator<Object> comparator = ((SortedMap)map).comparator();
                if (comparator != null) {
                    String string = "Cannot convert SortedMap with non-null comparator: " + comparator;
                    throw DefaultMXBeanMappingFactory.openDataException(string, new IllegalArgumentException(string));
                }
            }
            TabularType tabularType = (TabularType)this.getOpenType();
            TabularDataSupport tabularDataSupport = new TabularDataSupport(tabularType);
            CompositeType rowType = tabularType.getRowType();
            for (Map.Entry entry : map.entrySet()) {
                tabularDataSupport.put(new CompositeDataSupport(rowType, DefaultMXBeanMappingFactory.keyValueArray, new Object[] { this.keyMapping.toOpenValue(entry.getKey()), this.valueMapping.toOpenValue(entry.getValue()) }));
            }
            return tabularDataSupport;
        }
        
        @Override
        final Object fromNonNullOpenValue(Object o) throws InvalidObjectException {
            Collection<Object> collection = (Collection)Util.cast(((TabularData)o).values());
            Map<Object, Object> map = this.sortedMap ? Util.newSortedMap() : Util.newInsertionOrderMap();
            for (CompositeData compositeData : collection) {
                Object fromOpenValue = this.keyMapping.fromOpenValue(compositeData.get("key"));
                if (map.put(fromOpenValue, this.valueMapping.fromOpenValue(compositeData.get("value"))) != null) {
                    throw new InvalidObjectException("Duplicate entry in TabularData: key=" + fromOpenValue);
                }
                continue;
            }
            return map;
        }
        
        @Override
        public void checkReconstructible() throws InvalidObjectException {
            this.keyMapping.checkReconstructible();
            this.valueMapping.checkReconstructible();
        }
    }
    
    private final class CompositeMapping extends DefaultMXBeanMappingFactory.NonNullMXBeanMapping
    {
        private final String[] itemNames;
        private final Method[] getters;
        private final MXBeanMapping[] getterMappings;
        private DefaultMXBeanMappingFactory.CompositeBuilder compositeBuilder;
        
        CompositeMapping(Class<?> clazz, CompositeType compositeType, String[] itemNames, Method[] getters, MXBeanMappingFactory mxBeanMappingFactory) throws OpenDataException {
            super(clazz, compositeType);
            assert itemNames.length == getters.length;
            this.itemNames = itemNames;
            this.getters = getters;
            this.getterMappings = new MXBeanMapping[getters.length];
            for (int i = 0; i < getters.length; ++i) {
                this.getterMappings[i] = mxBeanMappingFactory.mappingForType(getters[i].getGenericReturnType(), mxBeanMappingFactory);
            }
            return;
        }
        
        @Override
        final Object toNonNullOpenValue(Object obj) throws OpenDataException {
            CompositeType compositeType = (CompositeType)this.getOpenType();
            if ((obj instanceof CompositeDataView)) {
                return ((CompositeDataView)obj).toCompositeData(compositeType);
            }
            if (obj == null) {
                return null;
            }
            Object[] itemValues = new Object[this.getters.length];
            for (int i = 0; i < this.getters.length; ++i) {
                try {
                    itemValues[i] = this.getterMappings[i].toOpenValue(MethodUtil.invoke(this.getters[i], obj, (Object[])null));
                }
                catch (Exception obj2) {
                    throw DefaultMXBeanMappingFactory.openDataException("Error calling getter for " + this.itemNames[i] + ": " + obj2, obj2);
                }
            }
            return new CompositeDataSupport(compositeType, this.itemNames, itemValues);
        }
        
        private synchronized void makeCompositeBuilder() throws InvalidObjectException {
            if (this.compositeBuilder != null) {
                return;
            }
            Class<Object> clazz = (Class)this.getJavaType();
            DefaultMXBeanMappingFactory.CompositeBuilder[][] array = { { new DefaultMXBeanMappingFactory.CompositeBuilderViaFrom(clazz, this.itemNames) }, { new DefaultMXBeanMappingFactory.CompositeBuilderViaConstructor(clazz, this.itemNames) }, { new DefaultMXBeanMappingFactory.CompositeBuilderCheckGetters(clazz, this.itemNames, this.getterMappings), new DefaultMXBeanMappingFactory.CompositeBuilderViaSetters(clazz, this.itemNames), new DefaultMXBeanMappingFactory.CompositeBuilderViaProxy(clazz, this.itemNames) } };
            DefaultMXBeanMappingFactory.CompositeBuilder compositeBuilder = null;
            StringBuilder obj = new StringBuilder();
            Throwable t = null;
        Label_0268:
            for (DefaultMXBeanMappingFactory.CompositeBuilder[] array2 : array) {
                int j = 0;
                while (j < array2.length) {
                    DefaultMXBeanMappingFactory.CompositeBuilder compositeBuilder2 = array2[j];
                    String applicable = compositeBuilder2.applicable(this.getters);
                    if (applicable == null) {
                        compositeBuilder = compositeBuilder2;
                        break Label_0268;
                    }
                    Throwable possibleCause = compositeBuilder2.possibleCause();
                    if (possibleCause != null) {
                        t = possibleCause;
                    }
                    if (applicable.length() > 0) {
                        if (obj.length() > 0) {
                            obj.append("; ");
                        }
                        obj.append(applicable);
                        if (j == 0) {
                            break;
                        }
                    }
                    ++j;
                }
            }
            if (compositeBuilder == null) {
                String str = "Do not know how to make a " + clazz.getName() + " from a CompositeData: " + obj;
                if (t != null) {
                    str += ". Remaining exceptions show a POSSIBLE cause.";
                }
                throw DefaultMXBeanMappingFactory.invalidObjectException(str, t);
            }
            else {
                this.compositeBuilder = compositeBuilder;
                return;
            }
        }
        
        @Override
        public void checkReconstructible() throws InvalidObjectException {
            this.makeCompositeBuilder();
        }
        
        @Override
        final Object fromNonNullOpenValue(Object o) throws InvalidObjectException {
            this.makeCompositeBuilder();
            return this.compositeBuilder.fromCompositeData((CompositeData)o, this.itemNames, this.getterMappings);
        }
    }
    
    private abstract static class CompositeBuilder
    {
        private final Class<?> targetClass;
        private final String[] itemNames;
        
        CompositeBuilder(Class<?> targetClass, String[] itemNames) {
            this.targetClass = targetClass;
            this.itemNames = itemNames;
        }
        
        Class<?> getTargetClass() {
            return this.targetClass;
        }
        
        String[] getItemNames() {
            return this.itemNames;
        }
        
        abstract String applicable(Method[] p0) throws InvalidObjectException;
        
        Throwable possibleCause() {
            return null;
        }
        
        abstract Object fromCompositeData(CompositeData p0, String[] p1, MXBeanMapping[] p2) throws InvalidObjectException;
    }
    
    private static final class CompositeBuilderViaFrom extends DefaultMXBeanMappingFactory.CompositeBuilder
    {
        private Method fromMethod;
        
        CompositeBuilderViaFrom(Class<?> clazz, String[] array) {
            super(clazz, array);
        }
        
        @Override
        String applicable(Method[] array) throws InvalidObjectException {
            Class<Object> targetClass = this.getTargetClass();
            try {
                Method method = targetClass.getMethod("from", new Class[] { CompositeData.class });
                if (!Modifier.isStatic(method.getModifiers())) {
                    throw new InvalidObjectException("Method from(CompositeData) is not static");
                }
                if (method.getReturnType() != this.getTargetClass()) {
                    throw new InvalidObjectException("Method from(CompositeData) returns " + MXBeanIntrospector.typeName(method.getReturnType()) + " not " + MXBeanIntrospector.typeName(targetClass));
                }
                this.fromMethod = method;
                return null;
            }
            catch (InvalidObjectException ex) {
                throw ex;
            }
            catch (Exception ex2) {
                return "no method from(CompositeData)";
            }
        }
        
        @Override
        final Object fromCompositeData(CompositeData compositeData, String[] array, MXBeanMapping[] array2) throws InvalidObjectException {
            try {
                return MethodUtil.invoke(this.fromMethod, null, new Object[] { compositeData });
            }
            catch (Exception ex) {
                throw DefaultMXBeanMappingFactory.invalidObjectException("Failed to invoke from(CompositeData)", ex);
            }
        }
    }
    
    private static class CompositeBuilderCheckGetters extends DefaultMXBeanMappingFactory.CompositeBuilder
    {
        private final MXBeanMapping[] getterConverters;
        private Throwable possibleCause;
        
        CompositeBuilderCheckGetters(Class<?> clazz, String[] array, MXBeanMapping[] getterConverters) {
            super(clazz, array);
            this.getterConverters = getterConverters;
        }
        
        @Override
        String applicable(Method[] array) {
            for (int i = 0; i < array.length; ++i) {
                try {
                    this.getterConverters[i].checkReconstructible();
                }
                catch (InvalidObjectException possibleCause) {
                    this.possibleCause = possibleCause;
                    return "method " + array[i].getName() + " returns type that cannot be mapped back from OpenData";
                }
            }
            return "";
        }
        
        @Override
        Throwable possibleCause() {
            return this.possibleCause;
        }
        
        @Override
        final Object fromCompositeData(CompositeData compositeData, String[] array, MXBeanMapping[] array2) {
            throw new Error();
        }
    }
    
    private static class CompositeBuilderViaSetters extends DefaultMXBeanMappingFactory.CompositeBuilder
    {
        private Method[] setters;
        
        CompositeBuilderViaSetters(Class<?> clazz, String[] array) {
            super(clazz, array);
        }
        
        @Override
        String applicable(Method[] array) {
            try {
                this.getTargetClass().getConstructor(new Class[0]);
            }
            catch (Exception ex) {
                return "does not have a public no-arg constructor";
            }
            Method[] setters = new Method[array.length];
            Method method;
            for (int i = 0; i < array.length; setters[i] = method, ++i) {
                Method obj = array[i];
                Class<Object> returnType = obj.getReturnType();
                String string = "set" + DefaultMXBeanMappingFactory.propertyName(obj);
                try {
                    method = this.getTargetClass().getMethod(string, new Class[] { returnType });
                    if (method.getReturnType() != Void.TYPE) {
                        throw new Exception();
                    }
                }
                catch (Exception ex2) {
                    return "not all getters have corresponding setters (" + obj + ")";
                }
            }
            this.setters = setters;
            return null;
        }
        
        @Override
        Object fromCompositeData(CompositeData compositeData, String[] array, MXBeanMapping[] array2) throws InvalidObjectException {
            Object instance;
            try {
                Class<Object> targetClass = this.getTargetClass();
                ReflectUtil.checkPackageAccess(targetClass);
                instance = targetClass.newInstance();
                for (int i = 0; i < array.length; ++i) {
                    if (compositeData.containsKey(array[i])) {
                        MethodUtil.invoke(this.setters[i], instance, new Object[] { array2[i].fromOpenValue(compositeData.get(array[i])) });
                    }
                }
            }
            catch (Exception ex) {
                throw DefaultMXBeanMappingFactory.invalidObjectException(ex);
            }
            return instance;
        }
    }
    
    private static final class CompositeBuilderViaConstructor extends DefaultMXBeanMappingFactory.CompositeBuilder
    {
        private List<DefaultMXBeanMappingFactory.CompositeBuilderViaConstructor.Constr> annotatedConstructors;
        
        CompositeBuilderViaConstructor(Class<?> clazz, String[] array) {
            super(clazz, array);
        }
        
        @Override
        String applicable(Method[] array) throws InvalidObjectException {
            if (!DefaultMXBeanMappingFactory.CompositeBuilderViaConstructor.AnnotationHelper.isAvailable()) {
                return "@ConstructorProperties annotation not available";
            }
            Constructor[] constructors = this.getTargetClass().getConstructors();
            List<Constructor> list = Util.newList();
            for (Constructor constructor : constructors) {
                if (Modifier.isPublic(constructor.getModifiers()) && DefaultMXBeanMappingFactory.CompositeBuilderViaConstructor.AnnotationHelper.getPropertyNames(constructor) != null) {
                    list.add(constructor);
                }
            }
            if (list.isEmpty()) {
                return "no constructor has @ConstructorProperties annotation";
            }
            this.annotatedConstructors = Util.newList();
            Map<Object, Integer> map = Util.newMap();
            String[] itemNames = this.getItemNames();
            for (int j = 0; j < itemNames.length; ++j) {
                map.put(itemNames[j], Integer.valueOf(j));
            }
            Set<Object> set = Util.newSet();
            for (Constructor constructor2 : list) {
                String[] propertyNames = DefaultMXBeanMappingFactory.CompositeBuilderViaConstructor.AnnotationHelper.getPropertyNames(constructor2);
                Type[] genericParameterTypes = constructor2.getGenericParameterTypes();
                if (genericParameterTypes.length != propertyNames.length) {
                    throw new InvalidObjectException("Number of constructor params does not match @ConstructorProperties annotation: " + constructor2);
                }
                int[] array2 = new int[array.length];
                for (int k = 0; k < array.length; ++k) {
                    array2[k] = -1;
                }
                BitSet o = new BitSet();
                int l = 0;
                while (l < propertyNames.length) {
                    String s = propertyNames[l];
                    if (!map.containsKey(s)) {
                        String s2 = "@ConstructorProperties includes name " + s + " which does not correspond to a property";
                        for (String str : map.keySet()) {
                            if (str.equalsIgnoreCase(s)) {
                                s2 = s2 + " (differs only in case from property " + str + ")";
                            }
                            else {
                                continue;
                            }
                        }
                        throw new InvalidObjectException(s2 + ": " + constructor2);
                    }
                    int intValue = (int)map.get(s);
                    array2[intValue] = l;
                    if (o.get(intValue)) {
                        throw new InvalidObjectException("@ConstructorProperties contains property " + s + " more than once: " + constructor2);
                    }
                    o.set(intValue);
                    Type genericReturnType = array[intValue].getGenericReturnType();
                    if (!genericReturnType.equals(genericParameterTypes[l])) {
                        throw new InvalidObjectException("@ConstructorProperties gives property " + s + " of type " + genericReturnType + " for parameter  of type " + genericParameterTypes[l] + ": " + constructor2);
                    }
                    ++l;
                }
                if (!set.add(o)) {
                    throw new InvalidObjectException("More than one constructor has a @ConstructorProperties annotation with this set of names: " + Arrays.toString(propertyNames));
                }
                this.annotatedConstructors.add(new DefaultMXBeanMappingFactory.CompositeBuilderViaConstructor.Constr(constructor2, array2, o));
            }
            for (BitSet set2 : set) {
                boolean n = false;
                for (BitSet set3 : set) {
                    if (set2 == set3) {
                        n = true;
                    }
                    else if (!(n == 0)) {
                        BitSet o2 = new BitSet();
                        o2.or(set2);
                        o2.or(set3);
                        if (!set.contains(o2)) {
                            TreeSet obj = new TreeSet();
                            for (int n2 = o2.nextSetBit(0); n2 >= 0; n2 = o2.nextSetBit(n2 + 1)) {
                                obj.add(itemNames[n2]);
                            }
                            throw new InvalidObjectException("Constructors with @ConstructorProperties annotation  would be ambiguous for these items: " + obj);
                        }
                        continue;
                    }
                    else {
                        continue;
                    }
                }
            }
            return null;
        }
        
        @Override
        final Object fromCompositeData(CompositeData compositeData, String[] array, MXBeanMapping[] array2) throws InvalidObjectException {
            CompositeType compositeType = compositeData.getCompositeType();
            BitSet set = new BitSet();
            for (int i = 0; i < array.length; ++i) {
                if (compositeType.getType(array[i]) != null) {
                    set.set(i);
                }
            }
            DefaultMXBeanMappingFactory.CompositeBuilderViaConstructor.Constr constr = null;
            for (DefaultMXBeanMappingFactory.CompositeBuilderViaConstructor.Constr constr2 : this.annotatedConstructors) {
                if (subset(constr2.presentParams, set) && (constr == null || subset(constr.presentParams, constr2.presentParams))) {
                    constr = constr2;
                }
                else {
                    continue;
                }
            }
            if (constr == null) {
                throw new InvalidObjectException("No constructor has a @ConstructorProperties for this set of items: " + compositeType.keySet());
            }
            Object[] initargs = new Object[constr.presentParams.cardinality()];
            for (int j = 0; j < array.length; ++j) {
                if (constr.presentParams.get(j)) {
                    Object fromOpenValue = array2[j].fromOpenValue(compositeData.get(array[j]));
                    int n = constr.paramIndexes[j];
                    if (n >= 0) {
                        initargs[n] = fromOpenValue;
                    }
                }
            }
            try {
                ReflectUtil.checkPackageAccess(constr.constructor.getDeclaringClass());
                return constr.constructor.newInstance(initargs);
            }
            catch (Exception ex) {
                throw DefaultMXBeanMappingFactory.invalidObjectException("Exception constructing " + this.getTargetClass().getName(), ex);
            }
        }
        
        private static boolean subset(BitSet set, BitSet set2) {
            BitSet set3 = (BitSet)set.clone();
            set3.andNot(set2);
            return set3.isEmpty();
        }
        
        static class AnnotationHelper
        {
            private static Class<? extends Annotation> constructorPropertiesClass;
            private static Method valueMethod;
            
            private static void findConstructorPropertiesClass() {
                try {
                    DefaultMXBeanMappingFactory.CompositeBuilderViaConstructor.AnnotationHelper.constructorPropertiesClass = Class.forName("java.beans.ConstructorProperties", false, DefaultMXBeanMappingFactory.class.getClassLoader());
                    DefaultMXBeanMappingFactory.CompositeBuilderViaConstructor.AnnotationHelper.valueMethod = DefaultMXBeanMappingFactory.CompositeBuilderViaConstructor.AnnotationHelper.constructorPropertiesClass.getMethod("value", new Class[0]);
                }
                catch (ClassNotFoundException ex) {}
                catch (NoSuchMethodException cause) {
                    throw new InternalError(cause);
                }
            }
            
            static boolean isAvailable() {
                return DefaultMXBeanMappingFactory.CompositeBuilderViaConstructor.AnnotationHelper.constructorPropertiesClass != null;
            }
            
            static String[] getPropertyNames(Constructor<?> constructor) {
                if (!isAvailable()) {
                    return null;
                }
                Annotation annotation = constructor.getAnnotation(DefaultMXBeanMappingFactory.CompositeBuilderViaConstructor.AnnotationHelper.constructorPropertiesClass);
                if (annotation == null) {
                    return null;
                }
                try {
                    return (String[])(String[])DefaultMXBeanMappingFactory.CompositeBuilderViaConstructor.AnnotationHelper.valueMethod.invoke(annotation, new Object[0]);
                }
                catch (InvocationTargetException cause) {
                    throw new InternalError(cause);
                }
                catch (IllegalAccessException cause2) {
                    throw new InternalError(cause2);
                }
            }
            
            static {
                findConstructorPropertiesClass();
            }
        }
        
        private static class Constr
        {
            final Constructor<?> constructor;
            final int[] paramIndexes;
            final BitSet presentParams;
            
            Constr(Constructor<?> constructor, int[] paramIndexes, BitSet presentParams) {
                this.constructor = constructor;
                this.paramIndexes = paramIndexes;
                this.presentParams = presentParams;
            }
        }
    }
    
    private static final class CompositeBuilderViaProxy extends DefaultMXBeanMappingFactory.CompositeBuilder
    {
        CompositeBuilderViaProxy(Class<?> clazz, String[] array) {
            super(clazz, array);
        }
        
        @Override
        String applicable(Method[] a) {
            Class<Object> targetClass = this.getTargetClass();
            if (!targetClass.isInterface()) {
                return "not an interface";
            }
            Set<Method> set = Util.newSet(Arrays.asList(targetClass.getMethods()));
            set.removeAll(Arrays.asList(a));
            String str = null;
            for (Method method : set) {
                String name = method.getName();
                Class[] parameterTypes = method.getParameterTypes();
                try {
                    if (!Modifier.isPublic(Object.class.getMethod(name, parameterTypes).getModifiers())) {
                        str = name;
                    }
                    else {
                        continue;
                    }
                }
                catch (NoSuchMethodException ex) {
                    str = name;
                }
            }
            if (str != null) {
                return "contains methods other than getters (" + str + ")";
            }
            return null;
        }
        
        @Override
        final Object fromCompositeData(CompositeData compositeData, String[] array, MXBeanMapping[] array2) {
            Class<Object> targetClass = this.getTargetClass();
            return Proxy.newProxyInstance(targetClass.getClassLoader(), new Class[] { targetClass }, new CompositeDataInvocationHandler(compositeData));
        }
    }
}
